
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_unsigned.all;
use ieee.numeric_std.all;

library unisim;
use unisim.vcomponents.all;

entity rapid_icap is
generic (
    C_BASEADDR   : std_logic_vector(0 to 31) := X"FFFFFFFF";
    C_MASK       : std_logic_vector(0 to 31) := X"00800000"
    );
  port (
    clk      : in  std_logic;
    rst      : in  std_logic;
    go  : in  std_logic;
    ready : out std_logic;
    BRAM_Rst : out std_logic;
    BRAM_Clk : out std_logic;
    BRAM_EN : out std_logic;
    BRAM_WEN : out std_logic_vector(0 to 3);
    BRAM_Addr : out std_logic_vector(0 to 31);
    BRAM_Din : in std_logic_vector(0 to 31);
    BRAM_Dout : out std_logic_vector(0 to 31)
    );
end entity rapid_icap ;

architecture IMP of rapid_icap is

  component pselect_mask is
    generic (
      C_AW   : integer                   := 32;
      C_BAR  : std_logic_vector(0 to 31) := X"00000000";
      C_MASK : std_logic_vector(0 to 31) := X"00800000");
    port (
      A     : in  std_logic_vector(0 to 31);
      CS    : out std_logic;
      Valid : in  std_logic);
  end component;
  
  
signal inc_cnt : std_logic:='0';
signal bs_done : std_logic:='0';
signal cnt: std_logic_vector(0 to 31);
signal ICAP_CE : std_logic;
signal enable : std_logic;
signal go_sig : std_logic;
signal bs_length : std_logic_vector(31 downto 0):=(others=>'0');
signal bs_length_read : std_logic:='0';

type state_type is (idle,read_bitstream_length,read_bitstream_length_1,process_dpr);
signal next_state, current_state: state_type;

signal addr_tmp : std_logic_vector(0 to 31):=(others => '0');
signal icap_datain : std_logic_vector(0 to 31);
signal BRAM_Din_tmp : std_logic_vector(0 to 31);
signal One : std_logic;
signal fsm_select : std_logic;
  
begin

BRAM_Addr <= addr_tmp;

BRAM_Rst <= rst;
BRAM_Clk <= clk;
go_sig <= go;

  SWAP_BITS: process(BRAM_Din)is
  begin
	for byte in 0 to 3 loop
		for bit in 0 to 7 loop
			icap_datain(byte*8+(7-bit)) <= BRAM_Din(byte*8 + bit);
		end loop; --Bit
	end loop; --Byte
  end process SWAP_BITS;


  Check_length: process (clk, rst, inc_cnt)
  begin
	 if (clk'event and clk = '1') then
      if ((rst = '1') or (bs_done = '1')) then
			cnt <= (others => '0');
			addr_tmp <= (others => '0');
      elsif inc_cnt = '1' then
			cnt <= cnt + 1;
			addr_tmp <= addr_tmp + 4;
		end if;
    end if;
  end process Check_length;

    read_bs_length: process (clk, rst, bs_length_read)
  begin
	 if (clk'event and clk = '1') then
      if (rst = '1') then
		bs_length <= (others => '0');
      elsif bs_length_read = '1' then
			bs_length <= BRAM_Din;
		end if;
    end if;
  end process read_bs_length;
  
  Reconf_done:process(clk, rst, bs_length, cnt)
  begin
  	 if (clk'event and clk = '1') then
      if rst = '1' then bs_done <= '0';
		elsif bs_length_read = '0' then
			if ((bs_length(31 downto 2)+1) = (cnt)) then bs_done <= '1';
			end if;
			else bs_done <= '0';
		end if;
	 end if;	
  end process Reconf_done;

  state_reg: process(clk, rst)
  begin
      if rst = '1' then
			current_state <= idle;
		elsif (clk'event and clk = '1') then
			current_state <= next_state;
		end if;
  end process state_reg;

  FSM: process(current_state, go_sig, bs_done)
  begin
  case current_state is
		when idle =>
						ICAP_CE <= '1';--Disable ICAP
						BRAM_EN <= '0';--Disable BRAM
						inc_cnt <= '0';
						ready <= '1';
						bs_length_read <= '0';
							if (go_sig='1') then next_state <= read_bitstream_length;
							else next_state <= idle;
							end if;

-- ENABLE BRAM
		when read_bitstream_length =>
						BRAM_EN <= '1';
						bs_length_read <= '0';
						inc_cnt <= '0';
						ready <= '0';
						next_state <= read_bitstream_length_1;

-- READ FIRST WORD (BITSTREAM SIZE)
		when read_bitstream_length_1 =>
						BRAM_EN <= '1';
						bs_length_read <= '1';
						inc_cnt <= '1';
						ready <= '0';
						next_state <= process_dpr;

-- PARTIAL RECONFIGURATION						
		when process_dpr =>
						bs_length_read <= '0';
						if (bs_done = '0') then
							next_state <= process_dpr;
							ICAP_CE <= '0';--Enable ICAP
							BRAM_EN <= '1';--Enable BRAM
							inc_cnt <= '1';
							ready <= '0';
						else	next_state <= idle;
							ICAP_CE <= '1';--Disable ICAP
							BRAM_EN <= '0';--Disable BRAM
							inc_cnt <= '0';
							ready <= '1';
						end if;
		when others =>
						bs_length_read <= '0';
						next_state <= idle;
						ICAP_CE <= '1';--Disable ICAP
						BRAM_EN <= '0';--Disable BRAM
						inc_cnt <= '0';
						ready <= '1';
	end case;
	end process FSM;

  pselect_mask_fsm : pselect_mask
    generic map (
      C_AW   => addr_tmp'length,
      C_BAR  => C_BASEADDR,
      C_MASK => C_MASK)
    port map (
      A     => addr_tmp,
      CS    => fsm_select,
      Valid => One);
	  
-- ICAP for Xilinx Virtex 5 familly

	ICAP_VERTEX5_I :  ICAP_VIRTEX5
	generic map(ICAP_WIDTH => "X32")
	port map (
	clk	=> clk,
	ce		=> ICAP_CE,
	write	=> '0',
	i		=> icap_datain,
	busy	=> open,
	o		=> open
	);

-- ICAP for Xilinx Virtex 6 familly
	
--		ICAP_VERTEX6_I :  ICAP_VIRTEX6
--	generic map(ICAP_WIDTH => "X32")
--	port map (
--	clk	=> clk,
--	csb		=> ICAP_CE,
--	rdwrb	=> '0',
--	i		=> icap_datain,
--	busy	=> open,
--	o		=> open
--	);
end architecture IMP;
